open Syntax
open Terms


type ('a, 'b) either = Left of 'a | Right of 'b

exception Unrefine
exception Not_Kind of string
exception Term_as_Type of string
exception Unreachable of string
exception SMT_Exec_Not_Found

let extensionality_mode = ref true

let debug = ref false

let count = ref 0
let fresh_var () =
  let res = "x__"^ string_of_int !count in
  incr count ; res

let fresh_tvar () =
  let res = "t__"^ string_of_int !count in
  incr count ; res

let fresh_vc_var () =
  let res = "vc__" ^ string_of_int !count in
  incr count ; res


let rec fresh_var_list = function
  | [] -> []
  | x::xs -> (x ^ fresh_var ()) :: fresh_var_list xs

(* Some context and kind uniformization *)

(** Performs a kind of "uncurrying" on Pi kinds, and so presuposes
    that k is of the form Pi x1:K1. .... Pi xn:Kn.K **)
let rec kpi_uncurry k =
  let rec uncurry_aux k acc =
    match k with
    | KPi(x,k1,k2) -> uncurry_aux k2 ((x,kpi_uncurry k1)::acc)
    | _ -> ("",k)::acc                     
    in
    match k with
    | KPi(_,_,_) -> let params = uncurry_aux k [] in
                     begin match params with
          | (_,k)::ks ->  KPiList(List.rev ks,k)
          | [] -> raise (Unreachable "uncurry")
                     end
    | _ -> k



let rec uncurry_app t =
  let rec uncurry_app_aux t acc =
    match t with
    | App(t1,t2) -> uncurry_app_aux t1 ((uncurry_app t2)::acc)
    | _ -> t::acc   in                   
  match t with
    | App(t1,t2) -> let params = uncurry_app_aux t1 [] in
        AppList(List.hd params, (List.tl params) @ [uncurry_app t2])
         
    | _ -> t 

(* f : (K1, … , Kn) -> {t::K | phi }

Faz com que for all fresh x1 :: K1 , … , xn:: Kn , phi(f (x1,…xn))

*)




let trivial_refine base_kind = KRefine( fresh_tvar() , base_kind ,  (Bool true) )




let unrefine k =
  match k with
  | KRefine(_,bk,_) -> bk
  | _ -> k

type prectxt = (string * term) list

type refctxt = (string * (term,term) either) list


let rec is_kind k =
  match k with
  | KRec | KFun | KType | KRef | KLab -> true
  | KGen(k') -> is_kind k'
  | KPi(_,k1,k2) -> is_kind k1 && is_kind k2
  | KPiList(l,k) ->  is_kind k && (List.for_all (fun (_,t) -> is_kind t) l)
  | KRefine(_,bk,f) -> is_kind bk && is_formula f
   | _ -> false
and
(* There is potential imprecision here due to overloading of record
   destructors *)
 is_type t =
  match t with
  Id(_) -> true
  | TInt | TBool | TUnit | 
  Ttop | Tbot | TEmptyRec | Label(_) | TString -> true
  | TRecCons(l,t1,t2) -> is_type l && is_type t1 && is_type t2
  | RLabel(t) -> is_type t
  | RHead(t) -> is_type t
  | RTail(t) -> is_type t
  | App(t1,t2) -> is_type t1 && is_type t2
  | AppList(t,l) -> is_type t && List.for_all is_type l
  | TArrow(t1,t2) -> is_type t1 && is_type t2
  | TDom(t) -> is_type t
  | TCodom(t) -> is_type t
  | TRef(t) -> is_type t
  | TRefOf(t) -> is_type t
  | LabSet(t) -> is_type t
  | LabConcat(l1,l2) -> is_type l1 && is_type l2
  | LabSetConcat(l,t) -> is_type l && is_type t                          
  | TIte(f,t1,t2) -> is_formula f &&  is_type t1 && is_type t2
  | _ -> false
and
  is_formula f =
  match f with
  | Id(_) -> true
  | Implies(f1,f2) -> is_formula f1 && is_formula f2
  | And(f1,f2) -> is_formula f1 && is_formula f2
  | Not(f) -> is_formula f
  | Equals(t1,t2,k) -> is_kind k && (is_type t1 || is_formula t1) &&
                       (is_type t1 || is_formula t2)
  | Or(f1,f2) -> is_formula f1 && is_formula f2
  | REmpty(t) -> is_type t

  | LabApart(t1,t2) -> is_type t1 && is_type t2
  | In(t1,t2) -> is_type t1 && is_type t2
  | Bool _ -> true
  | _ -> false

(* Takes a list of (string * term) and tags types with left and kinds with right, it also
makes it so that all kinds are refined or are Pi's of refined kinds -- trivially if needed. *)
let rec make_refctxt pctxt =
  match pctxt with
  | [] -> []
  | (x,t)::xs -> begin match t with
        TInt | TBool | TUnit | TString | 
        Ttop | Tbot | TEmptyRec |  TRecCons(_,_,_)
      | RLabel(_) | RHead(_) | RTail(_) | Label(_)
      | TForall(_,_,_) | TMap(_,_) | TArrow(_,_)
      | TDom(_) | TCodom(_) | TRef(_) | TRefOf(_) | TIte(_,_,_) |
      App(_,_) | Id(_) ->
        (x,Left (make_refctxt_aux t))::(make_refctxt xs)
      | KRec | KFun | KType | KRef | KLab | KGen(_) | KPi(_,_,_)
      | KRefine(_,_,_) -> (x, Right (higher_refine t))::(make_refctxt xs)
      | _ -> raise (Term_as_Type "make_refctxt")
        
       end
and higher_refine k =
  match k with
  | KRec | KFun | KType | KRef | KLab -> trivial_refine k
  | KGen(k) -> trivial_refine (KGen(higher_refine k))                                        
  | KPi(s,k1,k2) ->  higher_refine (kpi_uncurry k)
  | KPiList(l,k) -> KPiList(List.map (fun (s,k') -> (s,higher_refine k')) l, higher_refine k)
  | KRefine(s,bk,f) -> KRefine(s,bk, make_refctxt_aux f)
  | _ -> raise (Not_Kind "higher_refine")
and make_refctxt_aux t = (* DOES NOTHING *)
  match t with
  | _ -> t
    (*| Label(_) | TInt | TBool | TUnit |
      Ttop | Tbot | TEmptyRec | Bool _ | Id(_) -> t
    
    | TRecCons(t1,t2,t3) -> TRecCons(make_refctxt_aux t1,
                                     make_refctxt_aux t2,
                                     make_refctxt_aux t3)
    | RLabel(t) -> RLabel(make_refctxt_aux t)
    | RHead(t) -> RHead(make_refctxt_aux t)
    | RTail(t) -> RTail(make_refctxt_aux t)
    | TForall(s,t1,t2) -> TForall(s, make_refctxt_aux t1,
                                  make_refctxt_aux t2)
    | TMap(t1,t2) -> TMap(make_refctxt_aux t1, make_refctxt_aux t2)
    | TArrow(t1,t2) -> TArrow(make_refctxt_aux t1, make_refctxt_aux
                                t2)
    | TDom(t) -> TDom(make_refctxt_aux t)
    | TCodom(t) -> TCodom(make_refctxt_aux t)
    | TRefOf(t) -> TRefOf(make_refctxt_aux t)
    | Ite(f,t1,t2) -> Ite(make_refctxt_aux f, make_refctxt_aux t1, make_refctxt_aux t2)
    | TIte(f,t1,t2) -> TIte(make_refctxt_aux f, make_refctxt_aux t1,
                            make_refctxt_aux t2)
    | App(t1,t2) -> App(make_refctxt_aux t1, make_refctxt_aux t2) (*make_refctxt_aux (uncurry_app t)*)
    (*| AppList(t,l) -> AppList(t, List.map make_refctxt_aux l)*) (* I believe this is already taken care of *)

    | Implies(t1,t2) -> Implies(make_refctxt_aux t1, make_refctxt_aux
                                  t2)
    | And(t1,t2) -> And(make_refctxt_aux t1, make_refctxt_aux t2)
    | Or(t1,t2) -> Or(make_refctxt_aux t1, make_refctxt_aux t2)
    | Not(t) -> Not(make_refctxt_aux t)
    | Equiv(t1,t2,k) -> Equiv(make_refctxt_aux t1, make_refctxt_aux
                                t2, k)
    | Equals(t1,t2,k) -> Equals(make_refctxt_aux t1, make_refctxt_aux
                                  t2, k)
    | LabSet(t) -> LabSet(make_refctxt_aux t)
    | LabApart(t1,t2) -> LabApart(make_refctxt_aux t1,
                                  make_refctxt_aux t2)
    | REmpty(t) -> REmpty(make_refctxt_aux t)                           
    | In(t1,t2) -> In(make_refctxt_aux t1, make_refctxt_aux t2)

    | LabConcat(l1,l2) -> LabConcat(make_refctxt_aux l1,
                                    make_refctxt_aux l2)
    | LabSetConcat(l,t) -> LabSetConcat(make_refctxt_aux l,
                                        make_refctxt_aux t)
                              
      | _ -> raise (Unreachable "make_refctxt_aux")*)

(* Substitution *)
let rec subst env t =
  match t with
  (* Base homomorphic cases *)
  | KRec | KFun | KType | KRef | KLab | TString | String _ 
  | TInt | TBool | TUnit | Ttop | Tbot | TEmptyRec | Label(_) |
    Bool _ -> t
  (* Variable case *)
  | Id(s) -> begin match (Hashtbl.find_opt env s) with None -> Id(s)
                                                     | Some(t) -> t
             end
(* Formulas *)
  | Implies(t1,t2) -> Implies(subst env t1, subst env t2)                  
  | And(t1,t2) -> And(subst env t1, subst env t2)
  | Or(t1,t2) -> Or(subst env t1, subst env t2)
  | Equals(t1,t2,t3) -> Equals(subst env t1, subst env t2, subst env t3)
  | REmpty(t) -> REmpty(subst env t)
  | LabSet(t) -> LabSet(subst env t)
      | LabConcat(l1,l2) -> LabConcat(subst env l1,
                                    subst env l2)
    | LabSetConcat(l,t) -> LabSetConcat(subst env l,
                                        subst env t)
    | LabUnion(t1,t2) -> LabUnion(subst env t1, subst env t2)                             
  | LabApart(t1,t2) -> LabApart(subst env t1, subst env t2)
  | In(t1,t2) -> In(subst env t1, subst env t2)
  | Not(t) -> Not(subst env t)
(* Types *)
  | TRecCons(t1,t2,t3) -> TRecCons(subst env t1, subst env t2, subst env t3)
  | TArrow(t1,t2) -> TArrow(subst env t1, subst env t2)
  | TDom(t) -> TDom(subst env t)
  | TCodom(t) -> TCodom(subst env t)
  | TRef(t) -> TRef(subst env t)
  | TRefOf(t) -> TRefOf(t)
  | TIte(t1,t2,t3) -> TIte(subst env t1, subst env t2, subst env t3)
  | TLam(s,t1,t2) ->  let t1' = subst env t1 in
    Hashtbl.remove env s ; TLam(s,t1', subst env t2)
  | App(t1,t2) -> App(subst env t1, subst env t2)
  | AppList(t1,l) -> AppList(subst env t1, List.map (subst env) l)                    
  | TForall(s,t1,t2) ->  let t1' = subst env t1 in
         Hashtbl.remove env s ; TForall(s,t1',subst env t2)
  | TMap(t1,t2) -> TMap(subst env t1, subst env t2)
                     
  | RHead(t) -> RHead(subst env t)
  | RTail(t) -> RTail(subst env t)
  | RLabel(t) -> RLabel(subst env t)

(* Kinds *)
  | KGen(t) -> KGen(subst env t)
  | KPi(s,k1,k2) -> let k1' = subst env k1 in
    Hashtbl.remove env s ; KPi(s,k1',subst env k2)
  | KPiList(l,k) -> let l' = List.map (fun (s,k) -> (s,subst env k)) l in
    List.iter (fun (s,_) -> Hashtbl.remove env s) l ;
     KPiList(l',subst env k)
  | KRefine(s,k1,f) -> let k1' = subst env k1 in
    Hashtbl.remove env s ; KRefine(s,k1',subst env f)

  | _ -> raise (Term_as_Type "subst")
  
  
(***)



let rec encode t =
  (*print_term t ; print_endline " ENCODE " ; *)
 match t with
  (* Base cases *)
  | KRec | KFun | KType | KRef | KLab -> "Typ"
 

  (* Should these two encode to true and false? *)
  | Ttop -> "TRUE"           
  | Tbot -> "FALSE"
    
  | TUnit -> "Unit"
  | TInt -> "Num"
  | TBool -> "Bool"
  | Bool true -> "TRUE"
  | Bool false -> "FALSE" 
  | Label(s) -> "\"" ^ s ^ "\""
  | String s -> "\"" ^ s ^ "\""

  | TString -> "Str"
  (* Variable case *)
  | Id(s) -> s
(* Formulas *)
  | Implies(t1,t2) -> "(" ^ encode t1 ^ ") => (" ^ encode t2 ^ ")"     
  | And(t1,t2) ->  encode t1 ^ " AND " ^ encode t2
  | Or(t1,t2) ->  encode t1 ^ " OR " ^ encode t2 
  | Equals(t1,t2,t3) -> "("^encode t1 ^ " = " ^ encode t2^")"
  | EqVal(t1,t2) -> "("^encode t1 ^ " = " ^ encode t2^")"                        
  | REmpty(t) -> "is_nilrec(" ^ encode t ^ ")"
  | LabSet(t) -> "lab(" ^ encode t ^ ")"
  | LabApart(t1,t2) -> "apartl(" ^ encode t1 ^ "," ^ encode t2 ^ ")"
  | In(t1,t2) -> encode t1 ^ " IS_IN " ^ encode t2
  (*| Not(REmpty(t)) -> "is_cons(" ^ encode t ^ ")" *)
  | Not(t) -> "NOT (" ^ encode t ^")"
  
  (* Types *)
  | TEmptyRec -> "nilrec"
  | TRecCons(t1,t2,t3) -> "cons(" ^ encode t1 ^ "," ^ encode t2 ^ "," ^ encode t3 ^ ")"
  | TArrow(t1,t2) -> "FunType(" ^ encode t1 ^ "," ^ encode t2 ^ ")"
  | TDom(t) -> "dom(" ^ encode t ^")"
  | TCodom(t) ->"img(" ^ encode t ^")" 
  | TRef(t) -> "RefType(" ^ encode t ^ ")"
  | TRefOf(t) -> "refof(" ^ encode t ^ ")"
  | TIte(t1,t2,t3) -> "IF (" ^ encode t1 ^ ") THEN (" ^ encode t2 ^ ") ELSE (" ^ encode t3 ^ ") ENDIF"  
  | Ite(t1,t2,t3) -> "IF (" ^ encode t1 ^ ") THEN (" ^ encode t2 ^ ") ELSE (" ^ encode t3 ^ ") ENDIF"                     
  | TLam(s,t1,t2) ->  "nilrec"(*"LAMBDA (" ^ s ^ " : "^ encode t1 ^"): " ^ encode t2 *)
  | App(t1,t2) -> "appTyp(" ^ encode t1 ^ "," ^ encode t2 ^ ")"
  | AppList(t1,l) -> encode t1 ^ "(" ^ (String.concat "," (List.map encode l)) ^ ")"                  
  | TForall(s,t1,t2) ->  "PolyType(" ^ encode t2 ^")"
  | TMap(t1,t2) -> "map("  ^ encode t1 ^")"
                     
  | RHead(t) -> "ht(" ^ encode t ^")"
  | RTail(t) ->"tail(" ^ encode t ^")"
  | RLabel(t) ->"hlab(" ^ encode t ^")"
  | LabConcat(l1,l2) -> "CONCAT(" ^ encode l1 ^ " , " ^ encode l2 ^ ")"
  | LabSetConcat(l,t) -> "concatLab(" ^ encode l ^ "," ^ encode t ^ ")"
  | LabUnion(t1,t2) -> "(" ^ encode t1 ^ " | " ^ encode t2 ^ ")"
(* Kinds *)
  | KGen(t) -> "Typ"
  | KPi(s,k1,k2) -> "Typ" (* encode k1 ^ " -> " ^ encode k2 *)
  | KPiList(l,k) -> "Typ" (*"(" ^ String.concat "," (List.map (fun (_,_) ->
                      "Typ") l) ^ ") -> " ^ encode k     *)  
  | KRefine(s,k1,f) -> "Typ"

  | Lam(_,_,_) -> raise  (Term_as_Type "encode Lam")
  | RecF(f,_,_) -> f (* edited LC *)
(*  | RecF(_,_,_) -> raise  (Term_as_Type "encode RecF") *)
  | x -> ( print_term x; print_string " ??? \n\n"; raise (Term_as_Type "Xencode"))


let rec encode_bkind bk s =
  match bk with
  | KRec -> "is_rec(" ^ s ^ ")"
  | KFun -> "is_FunType(" ^ s ^ ")"
  | KType -> "TRUE"
  | KRef -> "is_RefType(" ^ s ^ ")"
  | KLab -> "TRUE"
  | KGen(t) -> "TRUE"
  | KPi(s,k1,k2) -> ""
  | KPiList(l,k2) -> ""           
  | KRefine(s,k1,f) -> encode_bkind k1 s
  | _ -> raise (Not_Kind "encode_bkind")
    


(** Function that collects all the necessary formulae from
   a ctx and generates it as a big conjunction **)
let rec gen_form_from_ctxt ctx = 
  match ctx with
  | [] -> []
  | (s,Left(t))::xs ->  (s ^ " = " ^ encode t) ::  gen_form_from_ctxt xs
  | (s,Right(KRefine(x,bk,Bool true)))::xs -> (encode_bkind bk s)::gen_form_from_ctxt xs
  | (s,Right(KRefine(x,bk,f)))::xs ->
    let env = Hashtbl.create 20 in
    let f' = Hashtbl.add env x (Id(s)) ; subst env f in
      (encode_bkind bk s)::(encode f')::gen_form_from_ctxt xs
  | (s,Right(KPi(s',k1,k2)))::xs -> gen_form_from_ctxt xs (* should no longer
                                                * occur *)
  | (s,Right(KPiList(l,k)))::xs -> gen_form_from_ctxt xs (* A bit hacky, but
                                 * conceptually simpler to separate *)
    
  | _ -> raise (Unreachable "gen_form_from_ctxt")
           
let gen_conj_from_ctxt ctx =                                 
    (gen_form_from_ctxt ctx) |> (String.concat " AND ")



(* f : Pi x:Type . {g:Fun | dom(g) = x} *)
(* f x :: {g:Fun | dom(g) = x} *)
(* dom(f x) = x *)

(* Special treatment of function signatures *)

let hkind_extract entry =
  (* Optionally produces a formula to be assumed *)
  let process_refinement k term vars fresh =
    match k with
    | KRefine(x,bk,Bool true) -> Some((encode_bkind bk (encode term),Bool true))
    | KRefine(x,bk,f) ->
      let env = Hashtbl.create 20 in
      let f' = Hashtbl.add env x term ;
        List.iter2 (fun v fv -> Hashtbl.add env v (Id(fv))) vars fresh ;
        subst env f in
      Some((encode_bkind bk (encode term),f'))
    | _ -> raise (Unreachable "process_refinement")
  in
  let gen_decls acc var k =
    match k with
    | KRefine(_,KLab,_) -> (var ^ " : STRING") :: acc
    | KLab -> (var ^ " : STRING") :: acc
    | _ -> (var ^ " : Typ") :: acc
  in
  let rec gen_defs l vars fresh kl =
    match (l,kl) with
    | (x::xs,(KRefine(s,bk,Bool true))::ks) ->
      (encode_bkind bk x)::(gen_defs xs vars fresh ks)
    | (x::xs,(KRefine(s,bk,f))::ks) ->
      let env = Hashtbl.create 20 in
      let f' = Hashtbl.add env s (Id(x)) ;
     List.iter2 (fun v fv -> Hashtbl.add env v (Id(fv))) vars fresh ; subst env f in
      (encode_bkind bk x)::(encode f')::gen_defs xs vars fresh ks
     | (x::xs,(KPiList(l,k))::ks) -> gen_defs xs vars fresh ks
    | ([],[]) -> []
    | _ -> raise (Unreachable "gen_defs")
  in
  match entry with
  | (s,Right(KPiList(l,k))) ->
    let split = List.split l in
    let vl = fst split in
    let kl = snd split in
    let fl = fresh_var_list vl in
    let app = List.fold_left (fun acc x -> App(acc,Id(x))) (Id(s)) fl    (*AppList(Id(s),(List.map (fun x -> Id x) fl)) in *) in
    let form = process_refinement k app vl fl in
     begin match form with
         None -> ("",[],[],None)
     | Some(_,Bool true) -> ("",[],[],None)
     | Some((s',f)) ->
       (* List of type declarations for the fresh variables *)
       let decls = List.rev @@ List.fold_left2 gen_decls [] fl kl in
       (* List of defining assertions for declarations *)
       let defs = gen_defs fl vl fl kl in
       (s,decls,defs,Some (s' ^ " AND " ^ encode f))
     end
    
    
  | _ -> ("",[],[],None)


(** Special treatment of Forall **)

(* Idea: Collect all the foralls and map them consistently to "fresh" terms,
   such that syntactically equal foralls map to the same term so that we can
   at least get equality **)

let rec eq_alpha t1 t2 env =
  match (t1,t2) with
  | (Id(x),Id(y)) -> (try (Hashtbl.find env x = Hashtbl.find env y)
                         with Not_found -> false)
  | (Num(n),Num(m)) -> n=m
  | (String(s),String(s')) -> s=s'
  | (TString,TString) -> true                             
  | (Bool(b),Bool(b')) -> b = b'
  | (Lam(x,t1,t2),Lam(y,s1,s2)) ->
    let x' = fresh_tvar () in
    eq_alpha t1 s1 env && (Hashtbl.add env x x' ; Hashtbl.add env y x' ;
                           eq_alpha t2 s2 env)
  | (App(t1,t2),App(s1,s2)) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env
  | (Clos(x,t1,e),Clos(y,t2,e')) ->
    let x' = fresh_tvar () in
    Hashtbl.add env x x' ; Hashtbl.add env y x' ; eq_alpha t1 t2 env
  | (FClos(x,t1,e),FClos(y,t2,e')) ->
    let x' = fresh_tvar () in
    Hashtbl.add env x x' ; Hashtbl.add env y x' ; eq_alpha t1 t2 env
  | (Ite(f,t1,t2),Ite(f',s1,s2)) ->
    eq_alpha f f' env && eq_alpha t1 s1 env && eq_alpha t2 s2 env
  | (Ref(t),Ref(s)) -> eq_alpha t s env
  | (Deref(t),Deref(s)) -> eq_alpha t s env
  | (Assign(t1,t2),Assign(s1,s2)) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env
  | (Loc(r),Loc(s)) -> eq_alpha !r !s env
  | (Kindcase(t1,t2,x,t3,t4),
     Kindcase(s1,s2,y,s3,s4)) ->
    let x' = fresh_tvar () in
    eq_alpha t1 s1 env && eq_alpha t2 s2 env &&
    eq_alpha t4 s4 env && (Hashtbl.add env x x' ; Hashtbl.add env y x' ; eq_alpha t3 s3 env)
  | (EmptyRec,EmptyRec) -> true
  | (RecCons(t1,t2,t3),RecCons(s1,s2,s3)) ->
    eq_alpha t1 s1 env && eq_alpha t2 s2 env && eq_alpha t3 s3 env
  | (RLabel(t),RLabel(s)) -> eq_alpha t s env
  | (RHead(t),RHead(s)) -> eq_alpha t s env
 | (RTail(t),RTail(s)) -> eq_alpha t s env
 | (REmpty(t),REmpty(s)) -> eq_alpha t s env
 | (TInt,TInt) -> true
 | (TBool,TBool) -> true
 | (TUnit,TUnit) -> true
 | (Ttop,Tbot) -> true
 | (TLam(x,t1,t2),TLam(y,s1,s2)) ->
   let x' = fresh_tvar () in
   eq_alpha t1 s1 env && (Hashtbl.add env x x' ; Hashtbl.add env y x' ; eq_alpha t2 s2 env)
 | (TClos(x,t,e),TClos(y,s,e')) ->
   let x' = fresh_tvar () in
   Hashtbl.add env x x' ; Hashtbl.add env y x' ; eq_alpha t s env
 | (TForall(x,t1,t2),TForall(y,s1,s2)) ->
   let x' = fresh_tvar () in
   eq_alpha t1 s1 env && (Hashtbl.add env x x' ; Hashtbl.add env y x' ; eq_alpha t2 s2 env)
 | (TMap(t1,t2),TMap(s1,s2)) ->
   eq_alpha t1 s1 env && eq_alpha t2 s2 env
 | (Label(s),Label(t)) -> s = t
 | TEmptyRec,TEmptyRec -> true
 | TRecCons(t1,t2,t3),TRecCons(s1,s2,s3) ->
   eq_alpha t1 s1 env && eq_alpha t2 s2 env && eq_alpha t3 s3 env
 | TArrow(t1,t2),TArrow(s1,s2) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env
 | TDom(t),TDom(s) -> eq_alpha t s env
 | TCodom(t),TCodom(s) -> eq_alpha t s env
 | TRef(t),TRef(s) -> eq_alpha t s env
 | TRefOf(t),TRefOf(s) -> eq_alpha t s env
 | TIte(f,t1,t2),TIte(f',s1,s2) ->
   eq_alpha f f' env && eq_alpha t1 s1 env && eq_alpha t2 s2 env
 | KRec,KRec -> true
 | KFun,KFun -> true
 | KType,KType -> true
 | KRef,KRef -> true
 | KLab,KLab -> true
 | KGen(t),KGen(s) -> eq_alpha t s env
 | KPi(x,t1,t2),KPi(y,s1,s2) ->
   let x' = fresh_tvar () in
   eq_alpha t1 s2 env && (Hashtbl.add env x x' ; Hashtbl.add env y x' ; eq_alpha t2 s2 env)
 | KRefine(x,t1,t2),KRefine(y,s1,s2) ->
   let x' = fresh_tvar () in
   eq_alpha t1 s2 env && (Hashtbl.add env x x' ; Hashtbl.add env y x' ; eq_alpha t2 s2 env)
 | Implies(t1,t2),Implies(s1,s2) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env
 | And(t1,t2),And(s1,s2) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env
  | Or(t1,t2),Or(s1,s2) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env
  | Not(t),Not(s) -> eq_alpha t s env
  | Equiv(t1,t2,t3),Equiv(s1,s2,s3) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env && eq_alpha t3 s3 env
  | Equals(t1,t2,t3),Equals(s1,s2,s3) ->  eq_alpha t1 s1 env && eq_alpha t2 s2 env && eq_alpha t3 s3 env
  | LabSet(t),LabSet(s) -> eq_alpha t s env
  | LabApart(t1,t2),LabApart(s1,s2) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env
  | In(t1,t2),In(s1,s2) -> eq_alpha t1 s1 env && eq_alpha t2 s2 env
  | LSet,LSet -> true
  | KindOK,KindOK -> true
  | KPiList(l,t),KPiList(l',s) -> List.for_all2 (fun (x,t) (y,s) ->
                                                let x' = fresh_tvar() in
                                                (Hashtbl.add env x x' ; Hashtbl.add env y x' ;
                                                eq_alpha t s env)) l l' && eq_alpha t s env 
  | AppList(t,l),AppList(s,l') ->
    eq_alpha t s env && List.for_all2 (fun t s -> eq_alpha t s env) l
      l'
  | LabConcat(l1,l2),LabConcat(l1',l2') -> eq_alpha l1 l1' env &&
  eq_alpha l2 l2' env
  | LabSetConcat(l1,t1),LabSetConcat(l2,t2) -> eq_alpha l1 l2 env &&
                                               eq_alpha t1 t2 env
  | LabUnion(t1,t2),LabUnion(t3,t4) -> eq_alpha t1 t3 env && eq_alpha t2 t4 env                                                
  | _ , _ -> false

    
module TermHash =
struct
  type t = term
  let equal t s = eq_alpha t s (Hashtbl.create 20)
  let hash = fun t -> 0 (* This is awful, but need a hash function compatible with eq_alpha... *)
end

module TermHashtbl = Hashtbl.Make(TermHash)


let mapcount = ref 0
let fresh_term_name () =
  let res = "forall__" ^ (string_of_int !mapcount) in
    incr mapcount ; res

let lookup table t =
  let o = TermHashtbl.find_opt table t in
  match o with
  | Some name -> Id(name)
  | None ->  let new_name = fresh_term_name () in
            TermHashtbl.add table t new_name ; Id(new_name)

let mk_decls table =
  TermHashtbl.fold
    (fun t v acc -> v ^ " : Typ;\nASSERT is_PolyType( " ^ v ^ ");\n" ^ acc) table ""

let collector env f =
  let rec pass table t =
    match t with
    | Id(_) | Num _ | Bool _ | String _ | TString
    | EmptyRec | TInt | TBool | TUnit | Ttop | Tbot
    | Label _ | TEmptyRec | KRec | KFun | KType
    | KRef | KLab | LSet | KindOK -> t
    | Lam(x,t1,t2) -> Lam(x,pass table t1,pass table t2)
    | App(t,s) -> App(pass table t, pass table s )
    | Ite(t1,t2,t3) -> Ite(pass table t1 , pass table t2 ,pass table t3)
    | Ref(t) ->Ref(pass table t)
    | Deref(t) ->Deref(pass table t)
    | Assign(t,s) ->Assign(pass table t ,pass table s)
    | Loc(rt) -> rt := pass table !rt ; t
    | Kindcase(t1,t2,x,t3,t4) -> Kindcase(pass table t1 , pass table t2 ,x , pass table t3 , pass table t4)
    | RecCons(t1,t2,t3) -> RecCons(pass table t1 , pass table t2 , pass table t3)
    | RLabel(t) -> RLabel(pass table t)
    | RHead(t) ->RHead(pass table t)
    | RTail(t) ->RTail(pass table t)
    | REmpty(t) ->REmpty(pass table t)
    | TLam(s,t1,t2) -> TLam(s,pass table t1 , pass table t2)
    | TForall(x,t1,t2) -> lookup table t
    | TMap(t1,t2) ->TMap(pass table t1 , pass table t2)
    | TRecCons(t1,t2,t3) -> TRecCons(pass table t1 , pass table t2 , pass table t3)
    | TArrow(s,t) -> TArrow(pass table s ,pass table t)
    | TDom(t) -> TDom(pass table t)
    | TCodom(t) ->TCodom(pass table t)
    | TRef(t) ->TRef(pass table t)
    | TRefOf(t) ->TRefOf(pass table t)
    | TIte(t1,t2,t3) ->TIte(pass table t1 , pass table t2 ,pass table t3)
    | KGen(t) -> KGen(pass table t)
    | KPi(s,t1,t2) ->KPi(s,pass table t1 , pass table t2)
    | KRefine(s,t1,t2) ->KRefine(s,pass table t1 , pass table t2)
    | Implies(t1,t2) -> Implies(pass table t1 ,pass table t2)
    | And(t1,t2) ->And(pass table t1 , pass table t2)
    | Not t -> Not(pass table t)
    | Equiv(t1,t2,t3) ->Equiv(pass table t1, pass table t2, pass table t3)
    | Equals(t1,t2,t3) ->Equals(pass table t1, pass table t2 , pass table t3)
    | Or(t1,t2) ->Or(pass table t1 , pass table t2)
    | LabSet(t) ->LabSet(pass table t)
    | LabApart(t1,t2) ->LabApart(pass table t1 ,pass table t2)
    | In(t1,t2) -> In(pass table t1 ,pass table t2)

    | KPiList(l,t) ->KPiList( List.map (fun (x,t) -> (x,pass table t)) l, pass table t)
    | AppList(t,l) -> AppList(pass table t , List.map (pass table) l)
    | LabConcat(l1,l2) -> LabConcat(pass table l1, pass table l2)
    | LabSetConcat(l,t) -> LabSetConcat(pass table l, pass table t)
    | LabUnion(l1,t) -> LabUnion(pass table l1, pass table t)                             
    | _ -> t
  in
  let table = TermHashtbl.create 20 in
  let env' = List.map (fun (s,t) -> (s,pass table t)) env in
  let f' = pass table f in
  let extra_decls = mk_decls table in
   (env',f',extra_decls)
  



(* Generating the type signature of a context *)


let ctxt_elem_enc elm acc =
  match elm with
  (* Produces the appropriate function signature *)
  | (s,Right(KPi(s',k1,k2))) ->
    (s ^ " : " ^ encode (KPi(s',k1,k2))) :: acc
  | (s,Right(KPiList(l,k))) ->
    (s ^ " : " ^ encode (KPiList(l,k))) :: acc
  (* Basic signatures *)
  | (s,Right(KRefine(_,KLab,_))) -> (s ^ " : STRING") :: acc
  | (s,Right(KLab)) -> (s ^ " : STRING") :: acc
  | (s,Left(_)) -> (s ^ " : Typ") :: acc
  | (s,Right(_)) -> (s ^ " : Typ") :: acc

let ctxt_as_tdecllist ctx = List.fold_right ctxt_elem_enc ctx []

(***)

let gen_higher_kind_data ctxt =
  let rec process_hkdata l ctxt   =
    match l with
    | (s,aux_decl,aux_defs,Some(f))::xs ->
      let def_s = List.hd (ctxt_elem_enc (s,(List.assoc s ctxt)) []) ^
                  ";\n" in
      let (p1,p2) = process_hkdata xs ctxt in
      (s::p1, def_s ^
      "ASSERT FORALL(" ^ (String.concat "," aux_decl) ^ "):\n("
      ^ (String.concat " AND " aux_defs) ^ ") => " ^ f ^ ";\n" ^ p2)
    | (_,_,_,None)::xs -> process_hkdata xs ctxt
    | [] -> ([],"")
  in
  let hkdata = List.map hkind_extract ctxt in
    process_hkdata hkdata ctxt


  
    


(* Encodes the check for |- T :: {s:K | f(s)} as the formula f(T) *)
let gen_vc_ref t rk =
  match rk with
  | KRefine(s,bk,f) -> let env = Hashtbl.create 20 in
    Hashtbl.add env s t ;
    let f' = subst env (make_refctxt_aux f) in
    let enc_bk = encode_bkind bk (encode t) in
    if (enc_bk = "TRUE") then encode f' else
            enc_bk ^ " AND " ^ encode f'
  | _ -> raise Unrefine


let gen_vc_formula f = encode (make_refctxt_aux f)
    
  

(* Encodes the subtyping check, assuming s1:[[bk1]] and s2:[[bk2]]
 * are in the surrounding context *)
let gen_vc_sub rk1 rk2 =
  match (rk1,rk2) with
  | (KRefine(_,_,f1),KRefine(_,_,f2)) ->
    encode (Implies(make_refctxt_aux f1, make_refctxt_aux f2))
  | _ -> raise (Unreachable "gen_vc_sub")
      
let is_first_order elm =
  match elm with
    (_,Left(_)) -> true
  | (_,Right(KPi(_,_,_))) -> false
  | (_,Right(KPiList(_,_))) -> false
  | (_,Right(_)) -> true

(* Encodes the check for Ctx |- T :: {s:K | f(s)} as 
QUERY FORALL (...): [[Ctx]] => [[f(T)]] *)
let gen_header_kinding ctx t rk =
  let (hk_vars,hk_data) = if (!extensionality_mode) then
      gen_higher_kind_data ctx else ([],"") in
  let fo_ctx = List.filter (fun (s,_) -> not (List.mem s hk_vars)) ctx in
  let enc =  (ctxt_as_tdecllist fo_ctx) |> (String.concat ",") in
  let ctxt_conj = gen_conj_from_ctxt ctx in
  let assump = if (ctxt_conj = "")
          then "TRUE" else ctxt_conj in
  let post = gen_vc_ref t rk in
  let forall = if (enc = "") then "" else "FORALL (" ^ enc ^
                                          "):\n" in
    hk_data ^ "QUERY " ^ forall ^ "(" ^ assump
   ^ ") => (" ^ post ^ ");"


(* Encodes the check for Ctx |- {s1:K1 | f(s1)} <: {s2:K2 | g(s2)} as 
[[Ctx]] && s1:[[K1]] && s2:[[K2]] ==> [[f(s1)]] ==> [[g(s2)]] *)
let gen_header_subkinding ctx rk1 rk2 =
  match (rk1,rk2) with
  | (KRefine(s1,bk1,f1),KRefine(s2,bk2,f2)) ->
   let (hk_vars,hk_data) = if (!extensionality_mode) then
      gen_higher_kind_data ctx else ([],"") in
  let fo_ctx = List.filter (fun (s,_) -> not (List.mem s hk_vars)) ctx in
  let ctx' = fo_ctx @ [(s1,Right(trivial_refine bk1));(s2,Right(trivial_refine bk2))] in 
  let enc =  (ctxt_as_tdecllist ctx') |> (String.concat " , ") in
  let ctxt_conj = gen_conj_from_ctxt ctx' in
  let assump = if (ctxt_conj = "") then
                encode (make_refctxt_aux f1) else ctxt_conj ^  " AND "
                                                    ^ encode (make_refctxt_aux f1) in
  let post = encode (make_refctxt_aux f2) in
  hk_data ^
  "QUERY FORALL (" ^ enc ^ "):\n    (" ^ assump
  ^ ") => (" ^ post ^ ");"
  | _ -> raise (Unreachable "gen_header_sub")


(* Some convenience top-level formatters *)
let gen_header_kinding_pc prectxt t rk =
 gen_header_kinding (make_refctxt prectxt) t rk

let gen_header_subkinding_pc prectxt rk1 rk2 =
  gen_header_subkinding (make_refctxt prectxt) rk1 rk2 

(***)


let rec gen_vcs_kinding l =
  match l with
  | [] -> []
  | (pctxt,t,rk)::xs -> (gen_header_kinding_pc pctxt t rk)::(gen_vcs_kinding xs)

let rec gen_vcs_subk l =
  match l with
  | [] -> []
  | (pctxt,rk1,rk2)::xs -> (gen_header_subkinding_pc pctxt rk1 rk2)::(gen_vcs_subk xs)


let header = "DATATYPE Typ = 
   nilrec  | cons(hlab:STRING, ht:Typ, tail:Typ)
  | FunType(dom:Typ, img:Typ) | Bool | Str
  | RefType(refof: Typ) | ColType(colof: Typ) | PolyType(map: Typ) | Unit | Num
END;
appTyp : (Typ,Typ) -> Typ;

is_rec : Typ -> BOOLEAN;
lab : Typ -> SET OF STRING;
apartl : (SET OF STRING, SET OF STRING) -> BOOLEAN;

ASSERT FORALL (r:Typ):
   is_rec(r) <=>  (is_nilrec(r) OR ( is_cons(r) AND is_rec(tail(r)) AND NOT(hlab(r) IS_IN lab(tail(r))))) ;

ASSERT FORALL (r:Typ): 
      ((is_rec(r) AND is_nilrec(r)) => (lab(r) = {} :: SET OF STRING)) AND
     ((is_rec(r) AND is_cons(r))  => ((lab(r) = {hlab(r)} | lab(tail(r))) )); 


ASSERT FORALL (s1:SET OF STRING, s2:SET OF STRING): 
(apartl (s1,s2) <=> FORALL (l:STRING): NOT ((l IS_IN s1) AND (l IS_IN s2)));


%apartl : (SET OF STRING, SET OF STRING) -> BOOLEAN;
ASSERT FORALL (s1:SET OF STRING, s2:SET OF STRING): 
(apartl (s1,s2) <=> (s1 & s2 = {} :: SET OF STRING));

%ASSERT FORALL (r1:Typ, r2:Typ): 
%(is_rec(r1) AND is_rec(r2) AND apartl(lab(r1),lab(r2)) AND NOT (is_nilrec(r1))) 
%=> apartl(lab(tail(r1)),lab(r2));
%ASSERT FORALL (r1:Typ, r2:Typ): 
%(is_rec(r1) AND is_rec(r2) AND apartl(lab(r1),lab(r2)) AND NOT (is_nilrec(r2))) 
%=> apartl(lab(r1),lab(tail(r2)));


concatLab :  (STRING, SET OF STRING) -> SET OF STRING;
ASSERT FORALL (s:STRING,sp:STRING,ss:SET OF STRING):
(s IS_IN ss) => (CONCAT(sp,s) IS_IN concatLab(sp,ss));

ASSERT FORALL (sp:STRING,s:STRING,ss:SET OF STRING):
(s IS_IN concatLab(sp,ss)) => EXISTS(suf:STRING): (suf IS_IN ss) AND (s=CONCAT(sp,suf));

%%% QUERIES BEGIN HERE

"
  


    


let gen_vcs l1 l2 =
  let kinding_vcs = gen_vcs_kinding l1 in
  let subk_vcs = gen_vcs_subk l2 in
  let k = if (kinding_vcs = []) then "" else
      String.concat "\n" ("PUSH;"::(kinding_vcs)@["POP;"]) in
  let sk = if (subk_vcs = []) then "" else
      String.concat "\n" ("PUSH;"::(subk_vcs)@["POP;"]) in
  header ^ k ^ "\n" ^ sk ^ "\n"






let write_vc_file fname vc =
  let fhandler = open_out_bin fname in
  Printf.fprintf fhandler "%s" vc ; close_out fhandler

(** One shot run of cvc4 process to test for a given VC **)
let run_cvc4 vc =
  let cmd = "cvc4 --lang cvc4 --tlimit-per=2000 --incremental -q" in
  begin match (Unix.open_process cmd) with
    | (inchan,outchan) ->
      output_string outchan vc ;
      close_out outchan ;
      let lines = ref [] in
      try
        while true do
          lines := input_line inchan :: !lines
        done; !lines
      with End_of_file ->
        close_in inchan;
        List.rev !lines
end

let parse_cvc_out lines =
  List.for_all (fun s -> s = "valid") lines


(** Interactive CVC IO **)
let cvc_in = ref None
let cvc_out = ref None
let cvc_err = ref None

(** Launches the CVC process and grabs the IO channels **)
let init_cvc_proc () =
  let cmd = "cvc4 --lang cvc4 --tlimit-per=1000 --incremental -q" in
  begin match (Unix.open_process_full cmd (Unix.environment ()) ) with
    | (inchan,outchan,error) ->
      cvc_in := Some(inchan) ;
      cvc_out := Some(outchan) ;
      cvc_err := Some(error) ;
      close_in error ;
      output_string outchan header ;
      flush outchan
  end

let clean_up_cvc_proc () =
  match (!cvc_in,!cvc_out,!cvc_err) with
  | (None,None,None) -> ()
  | (Some(inc),Some(outc),Some(err) ) ->
    let _ = Unix.close_process_full (inc,outc,err) in
    cvc_in := None ;
    cvc_out := None ;
    cvc_err := None 
  | _ -> raise (Unreachable "clean_up")


let validity_check () =
  match (!cvc_out,!cvc_in) with
  | (Some(outc),Some(inc)) ->
     output_string outc "PUSH;\nQUERY FALSE;\nPOP;\n" ;
     flush outc ;
     (try 
        let ans = input_line inc in
 (*       if (!debug) then
            print_endline ans ;*)
     (ans<>"valid")
      with End_of_file ->
        clean_up_cvc_proc () ; init_cvc_proc () ; true)
  | _ -> raise SMT_Exec_Not_Found

let make_query q =
  if (!debug) then
    (print_endline "QUERY BEGIN" ; print_endline q ;  print_endline "QUERY END" );
  match (!cvc_out,!cvc_in) with
  |(Some(outc),Some(inc)) ->
  (* print_endline ("Validity check: "^ string_of_bool (validity_check
   * ())) ; *)
  output_string outc q ;
  flush outc ;
  (try 
     let ans = input_line inc in
     (if (!debug) then
       print_endline ("ANSWER: " ^ ans));
     (if (ans = "unknown (TIMEOUT)") then
        (clean_up_cvc_proc () ; init_cvc_proc ()));
    (ans="valid")
  with End_of_file ->
    clean_up_cvc_proc () ; init_cvc_proc () ;
    if (!debug) then print_endline "EOF Reached" ; false)
  | _ -> raise SMT_Exec_Not_Found



(* Some incremental encoding utils *)
let new_ctxt_elem x =
  let ctxt = make_refctxt [x] in
    let (hk_vars,hk_data) = if (!extensionality_mode) then
      gen_higher_kind_data ctxt else ([],"") in
  let fo_ctx = List.filter (fun (s,_) -> not (List.mem s hk_vars)) ctxt in
  let decls = ctxt_as_tdecllist fo_ctx in
  let formulas = gen_form_from_ctxt fo_ctx in
  (hk_data,decls,formulas)

let prep_query extra_decls hk_data decls assumps post =
  let decl_string = String.concat "," decls in
  let assump_string = String.concat " AND " assumps in 
  let forall = if (decl_string = "") then "" else "FORALL (" ^ decl_string ^
                                                  "):\n" in
  let body = if (assump_string = "") then post ^ ";\nPOP;\n" else "(" ^ assump_string ^
  ") => (" ^ post ^ ");\nPOP;\n" in 
                
  "PUSH;\n" ^ extra_decls ^ hk_data ^ "\n QUERY " ^ forall ^ body 

(***)





(* Tests *)

let test_ctxt0 = [("x1",KPi("a",KFun,
                            KPi("b",KFun,
                                KPi("c",KType,
                                    KRefine("d",KType,
                                            And(Equals(Id("d"),TDom(Id("a")),KType),
                                                Equals(Id("d"),TCodom(Id("b")),KType)))))))]



let test_ctxt1 = [("x2",KPi("a",KRefine("e",KFun,Ttop),
                            KPi("b",KFun,
                                KPi("c",KRefine("f",KType,Equals(Id("f"),TBool,KType)),
                                    KRefine("d",KType,
                                            Equals(Id("d"),Id("c"),KType)
                                                )))))]

let test_ctxt2 = [("x3",KPi("a",(KPi("anon",KType,KType)),
                            KPi("b",KType,
                                KRefine("d",KType,
                                        Equals(App(Id("a"),Id("b")),Id("d"), KType)))))]
                          
let test_ctxt3 = [("x4",KPi("a",KFun,
                            KPi("b",KRefine("b0",KType,Equals(Id("b0"),TDom(Id("a")),KType)),
                                KRefine("d",KType,Equals(Id("d"),Id("b"),KType)))))]


let test_tm0 = KRefine("x",KType,Equals(Id("x"),TBool,Ttop))


let test_ctxt4 = [("x1",KPi("a",KType,
                            KRefine("b",KFun,Equals(TDom(Id("b")),Id("a"),KType))))]
let test_ctxt5 = [("x1",KPi("a",KType,
                            KRefine("b",KFun,Equals(TDom(Id("b")),Id("a"),KType))));
                  ("x2",KPi("a",KType,
                            KRefine("b",KFun,And(Equals(TDom(Id("b")),Id("a"),KType),
                                                 Equals(TCodom(Id("b")),Id("a"),KType)))))]                 
let test_type1 = App(Id("x1"),TBool)
let test_type2 = App(Id("x2"),TBool)
    
let test_kind1 = KRefine("c",KFun,Equals(TDom(Id("c")),TBool,KType))
let test_kind2 = KRefine("c",KFun,Equals(TCodom(Id("c")),TBool,KType))

(* Dont forget to build with unix.cma  *)

(*


DATATYPE Typ = 
   nilrec  | cons(hlab:STRING, ht:Typ, tail:Typ)
  | FunType(dom:Typ, img:Typ) | Bool 
  | RefType(refof: Typ) | ColType(colof: Typ) | PolyType(map: Typ) | Unit | Num
END;
appTyp : (Typ,Typ) -> Typ;
appTyp2 : (Typ,Typ,Typ) -> Typ;

is_rec : Typ -> BOOLEAN;
lab : Typ -> SET OF STRING;
apartl : (SET OF STRING, SET OF STRING) -> BOOLEAN;

ASSERT FORALL (r:Typ):
   is_rec(r) <=>  (is_nilrec(r) OR ( is_cons(r) AND is_rec(tail(r)) AND NOT(hlab(r) IS_IN lab(tail(r))))) ;

ASSERT FORALL (r:Typ): 
      ((is_rec(r) AND is_nilrec(r)) => (lab(r) = {} :: SET OF STRING)) AND
     ((is_rec(r) AND is_cons(r))  => ((lab(r) = {hlab(r)} | lab(tail(r))) )); 


ASSERT FORALL (s1:SET OF STRING, s2:SET OF STRING): 
(apartl (s1,s2) <=> FORALL (l:STRING): NOT ((l IS_IN s1) AND (l IS_IN s2)));


%apartl : (SET OF STRING, SET OF STRING) -> BOOLEAN;
%ASSERT FORALL (s1:SET OF STRING, s2:SET OF STRING): 
%(apartl (s1,s2) <=> (s1 & s2 = {} :: SET OF STRING));

ASSERT FORALL (r1:Typ, r2:Typ): 
(is_rec(r1) AND is_rec(r2) AND apartl(lab(r1),lab(r2)) AND NOT (is_nilrec(r1))) 
=> apartl(lab(tail(r1)),lab(r2));
ASSERT FORALL (r1:Typ, r2:Typ): 
(is_rec(r1) AND is_rec(r2) AND apartl(lab(r1),lab(r2)) AND NOT (is_nilrec(r2))) 
=> apartl(lab(r1),lab(tail(r2)));


concatLab :  (STRING, SET OF STRING) -> SET OF STRING;
ASSERT FORALL (s:STRING,sp:STRING,ss:SET OF STRING):
(s IS_IN ss) => (CONCAT(sp,s) IS_IN concatLab(sp,ss));

ASSERT FORALL (sp:STRING,s:STRING,ss:SET OF STRING):
(s IS_IN concatLab(sp,ss)) => EXISTS(suf:STRING): (suf IS_IN ss) AND (s=CONCAT(sp,suf));

%%% QUERIES BEGIN HERE

RAppend : Typ;
ASSERT FORALL(RAx__56 : Typ,RBx__55 : Typ):
(is_rec(RAx__56) AND is_rec(RBx__55) AND apartl(lab(RBx__55),lab(RAx__56))) => is_rec(appTyp(appTyp(RAppend,RAx__56),RBx__55)) AND 
(lab(appTyp(appTyp(RAppend,RAx__56),RBx__55)) = (lab(RAx__56) | lab(RBx__55)));

ASSERT FORALL(RAx__56 : Typ,RBx__55 : Typ):
(is_rec(RAx__56) AND is_rec(RBx__55) AND apartl(lab(RBx__55),lab(RAx__56))) => is_rec(appTyp2(RAppend,RAx__56,RBx__55)) AND 
(lab(appTyp2(RAppend,RAx__56,RBx__55)) = (lab(RAx__56) | lab(RBx__55)));

 QUERY FORALL (X$12 : Typ,X$7 : Typ,X$1 : Typ):
(is_cons(X$1) AND is_rec(X$7) AND apartl(lab(X$7),lab(X$1)) AND is_rec(X$1)) => (NOT (hlab(X$1) IS_IN lab(appTyp2(RAppend,tail(X$1),X$7))));


 QUERY FORALL (X$12 : Typ,X$7 : Typ,X$1 : Typ):
(is_rec(X$1) AND is_cons(X$1) AND is_rec(X$7) AND apartl(lab(X$7),lab(X$1)) AND is_rec(X$1)) => (NOT (hlab(X$1) IS_IN lab(tail(X$1))));

 QUERY FORALL (X$12 : Typ,X$7 : Typ,X$1 : Typ):
( is_cons(X$1) AND is_rec(X$7) AND apartl(lab(X$7),lab(X$1)) AND is_rec(X$1)) => (NOT (hlab(X$1) IS_IN lab(X$7)));

 QUERY FORALL (X$12 : Typ,X$7 : Typ,X$1 : Typ):
( is_cons(X$1) AND is_rec(X$7) AND apartl(lab(X$7),lab(X$1)) AND is_rec(X$1)) => is_rec(appTyp(appTyp(RAppend,X$1),X$7));


 QUERY FORALL (X$12 : Typ,X$7 : Typ,X$1 : Typ):
( is_cons(X$1) AND is_rec(X$7) AND apartl(lab(X$7),lab(X$1)) AND is_rec(X$1)) => (lab(appTyp(appTyp(RAppend,tail(X$1)),X$7)) = (lab(tail(X$1)) | lab(X$7)));

 QUERY FORALL (X$7 : Typ,X$1 : Typ):
(is_cons(X$1) AND is_rec(X$7) AND apartl(lab(X$7),lab(X$1)) AND is_rec(X$1)) => (NOT (hlab(X$1) IS_IN lab(X$7))) AND (NOT (hlab(X$1) IS_IN (lab(tail(X$1)) | lab(X$7))));

 QUERY FORALL (X$12 : Typ,X$7 : Typ,X$1 : Typ):
(is_cons(X$1) AND is_rec(X$7) AND apartl(lab(X$7),lab(X$1)) AND is_rec(X$1)) => (NOT (hlab(X$1) IS_IN lab(appTyp(appTyp(RAppend,tail(X$1)),X$7))));

 QUERY FORALL (X$12 : STRING,X$7 : SET OF STRING ,X$1 : SET OF STRING ):
(NOT (X$12 IS_IN X$7)) AND  (NOT (X$12 IS_IN X$1)) => (NOT (X$12 IS_IN (X$1 | X$7))) ;

*)
