
open Syntax
open Terms


type ('a, 'b) either = Left of 'a | Right of 'b

exception Unrefine
exception Not_Kind of string
exception Term_as_Type of string
exception Unreachable of string

(* Flag that controls whether or not to apply extensionality to higher
 * kinds *)
val extensionality_mode : bool ref
val debug : bool ref

(* Helper functions *)

type refctxt = (string * (term,term) either) list
           

(** trivial_refine k = KRefine(s,k,true), where s is a pseudo-fresh symbol. 
   Assumes that k is a base kind. *)
val trivial_refine : term -> term

(** Transforms a termenv into a ctxt that is manipulated by the VC Generator
   module *)
val make_refctxt : termenv -> refctxt

(** Testers -- very basic shape analysis, no 
 * sophisticated well-formedness checks *)
val is_kind : term -> bool
val is_type : term -> bool
val is_formula : term -> bool



(* Main functions *)

val encode:  term -> string

val encode_bkind : term -> string -> string

(** gen_vc_ref t rk generates a string encoding of the underlying
    refinement of rk, with t instantiating the formal parameter **)
val gen_vc_ref : term -> term -> string


(** gen_vc_formula f generates a string that can be fed
   to cvc4 that encodes the check for validity of f *)
val gen_vc_formula : term -> string

(** gen_vcs l1 l2 generates a string to be fed to run_cvc4, where
   l1 is a list of elements of the form (ctxt,T,KRefine(s,K,F)) and 
   l2 is a list of elements of the form (ctxt,KRefine(s,K1,F1),KRefine(s,K2,F2)). 
   l1 will be used to generate verification conditions of the form
   [[Ctxt]] => [[F(T)]] , and l2 will be used to generate VCs of the
   form [[Ctxt]] => [[F1]] => [[F2]]. *)
val gen_vcs : (termenv * term * term) list -> (termenv * term * term) list -> string

(** write_vc_file fname vc writes to fname the vc, assuming
    vc was generated by gen_vcs **)
val write_vc_file : string -> string -> unit

(** Given a file as a string, run cvc4 on it and produce a list of output strings *)
val run_cvc4 : string -> string list


(** Takes the cvc4 output and returns true iff all VCs are valid *)
val parse_cvc_out : string list -> bool


(** Interactive CVC usage **)

(** Initiates the CVC process **)
val init_cvc_proc : unit -> unit
(** Terminates the process **)
val clean_up_cvc_proc : unit -> unit

(** Helper to allow for incremental context encoding generation,
    new_ctxt_elem ctxt_element produces (hk_data,decls,assumps), 
    where hk_data is the encoded higher-kind data, decls is a
    declaration list (to be FORALL'd) and assumps are formulas that
    are to be assumed.  **)
val new_ctxt_elem : (string * term) -> (string * string list * string list)

(** Specialized functions to handle forall **)


val collector : termenv -> term -> (termenv * term * string)



(** Prepares an individual query string, given the data generated from collector/mk_decls, the 3 data generated by
    new_ctxt_elem and a formula **)
val prep_query : string -> string -> string list -> string list -> string -> string

(** Performs the given query by interacting with a running CVC process **)
val make_query : string -> bool
  
(** Asks whether false is not provable, i.e. QUERY FALSE <> valid **)
val validity_check : unit -> bool

